/*****************************************************************************
 * DeDRMS.cs: DeDRMS 0.4
 *****************************************************************************
 * Copyright (C) 2004 Jon Lech Johansen <jon-vl@nanocrew.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

using System;
using System.IO;
using System.Text;
using System.Security.Cryptography;

class M4PStream
{
    private Rijndael alg;

    private BinaryReader br;
    private BinaryWriter bw;
    private byte [] sbuffer;

    private string AtomDRMS = "drms";
    private string AtomMP4A = "mp4a";
    private string AtomSINF = "sinf";
    private string AtomUSER = "user";
    private string AtomKEY  = "key ";
    private string AtomIVIV = "iviv";
    private string AtomNAME = "name";
    private string AtomPRIV = "priv";
    private string AtomSTSZ = "stsz";
    private string AtomSTCO = "stco";
    private string AtomCO64 = "co64";
    private string AtomSTSC = "stsc";
    private string AtomGEID = "geID";

    public M4PStream( FileStream fs )
    {
        br = new BinaryReader( fs );
        bw = new BinaryWriter( fs );
        sbuffer = br.ReadBytes( Convert.ToInt32( fs.Length ) );

        alg = Rijndael.Create();
        alg.Mode = CipherMode.CBC;
        alg.Padding = PaddingMode.None;
    }

    private byte [] LeReverse( byte [] Input, int Pos, int Count )
    {
        if( BitConverter.IsLittleEndian )
        {
            for( int i = 0; i < Count; i++ )
            {
                Array.Reverse( Input, Pos + (i * 4), 4 );
            }
        }

        return Input;
    }

    private int GetAtomPos( string Atom, bool ThrowException )
    {
        byte [] Bytes = Encoding.ASCII.GetBytes( Atom );

        for( int i = 0; i < (sbuffer.Length - 3); i++ )
        {
            if( sbuffer[ i + 0 ] == Bytes[ 0 ] &&
                sbuffer[ i + 1 ] == Bytes[ 1 ] &&
                sbuffer[ i + 2 ] == Bytes[ 2 ] &&
                sbuffer[ i + 3 ] == Bytes[ 3 ] )
            {
                return i;
            }
        }

        if( !ThrowException )
        {
            return -1;
        }

        throw new Exception( String.Format( "Atom '{0}' not found", Atom ) );
    }

    private uint GetAtomSize( int Pos )
    {
        byte [] Bytes = new byte[ 4 ];
        Buffer.BlockCopy( sbuffer, Pos - 4, Bytes, 0, 4 );
        return BitConverter.ToUInt32( LeReverse( Bytes, 0, 1 ), 0 );
    }

    private byte [] GetAtomData( int Pos, bool bLeReverse )
    {
        uint Size;
        byte [] Bytes;

        Size = GetAtomSize( Pos );
        Bytes = new byte[ Size - 8 ];
        Buffer.BlockCopy( sbuffer, Pos + 4, Bytes, 0, Bytes.Length );

        return bLeReverse ? LeReverse( Bytes, 0, Bytes.Length / 4 ) : Bytes;
    }

    private void Decrypt( byte [] CipherText, int Offset, int Count,
                          byte [] Key, byte [] IV )
    {
        MemoryStream ms = new MemoryStream();

        ICryptoTransform ct = alg.CreateDecryptor( Key, IV );
        CryptoStream cs = new CryptoStream( ms, ct, CryptoStreamMode.Write );
        cs.Write( CipherText, Offset, (Count / 16) * 16 );
        cs.Close();

        ms.ToArray().CopyTo( CipherText, Offset );
    }

    private byte [] GetUserKey( uint UserID, uint KeyID )
    {
        byte [] UserKey;
        BinaryReader bruk;

        bool bUnix = Environment.OSVersion.ToString().IndexOf( "Unix" ) != -1;
        string strHome = bUnix ? Environment.GetEnvironmentVariable( "HOME" ) :
            Environment.GetFolderPath( Environment.SpecialFolder.ApplicationData );
        string strFile = String.Format( "{0}{1}{2}drms{3}{4:X8}.{5:D3}", strHome,
                                        Path.DirectorySeparatorChar, bUnix ? "." : "",
                                        Path.DirectorySeparatorChar, UserID, KeyID );

        // ADDED(2): iOpener
		try
		{
			bruk = new BinaryReader( File.OpenRead( strFile ) );
		// ADDED(5): iOpener
		}
		catch (IOException ioe) 
		{
			throw new iOpener.KeyNotFoundException();
		}

        UserKey = bruk.ReadBytes( Convert.ToInt32( bruk.BaseStream.Length ) );
        bruk.Close();

        return UserKey;
    }

    private int [] GetSampleSizeTable()
    {
        byte [] adSTSZ = GetAtomData( GetAtomPos( AtomSTSZ, true ), true );
        int Count = BitConverter.ToInt32( adSTSZ, 8 );
        int [] Table = new int[ Count ];

        for( int i = 0; i < Count; i++ )
        {
            Table[ i ] = BitConverter.ToInt32( adSTSZ, 12 + (i * 4) );
        }

        return Table;
    }

    private int [] GetChunkOffsetTable()
    {
        byte [] adCO64;
        int apSTCO = GetAtomPos( AtomSTCO, false );

        if( apSTCO != -1 )
        {
            adCO64 = GetAtomData( apSTCO, true );
        }
        else
        {
            adCO64 = GetAtomData( GetAtomPos( AtomCO64, true ), true );
        }

        int Count = BitConverter.ToInt32( adCO64, 4 );
        int [] Table = new int[ Count ];

        for( int i = 0; i < Count; i++ )
        {
            if( apSTCO != -1 )
            {
                Table[ i ] = BitConverter.ToInt32( adCO64, 8 + (i * 4) );
            }
            else
            {
                Table[ i ] = Convert.ToInt32(
                    BitConverter.ToInt64( adCO64, 8 + (i * 8) ) );
            }
        }

        return Table;
    }

    private int [] GetChunkSamplesTable( int ChunkCount )
    {
        byte [] adSTSC = GetAtomData( GetAtomPos( AtomSTSC, true ), true );
        int Count = BitConverter.ToInt32( adSTSC, 4 );
        int [] FirstChunk = new int[ Count ];
        int [] SamplesPerChunk = new int[ Count ];

        for( int i = 0; i < Count; i++ )
        {
            FirstChunk[ i ] = BitConverter.ToInt32( adSTSC, 8 + (i * 12) );
            SamplesPerChunk[ i ] = BitConverter.ToInt32( adSTSC, 12 + (i * 12) );
        }

        int [] Table = new int[ ChunkCount ];
        int Last = ChunkCount;
        int Index = Count;

        while( Index-- > 0 )
        {
            for( int Chunk = FirstChunk[ Index ] - 1;
                 Chunk < Last; Chunk++ )
            {
                Table[ Chunk ] = SamplesPerChunk[ Index ];
            }

            Last = FirstChunk[ Index ] - 1;
        }

        return Table;
    }

    public void DeDRMS()
    {
        byte [] IV = new byte[ 16 ];
        byte [] Key = new byte[ 16 ];

        int apDRMS = GetAtomPos( AtomDRMS, true );
        int apSINF = GetAtomPos( AtomSINF, true );
        int apGEID = GetAtomPos( AtomGEID, true );

        uint asDRMS = GetAtomSize( apDRMS );
        uint asSINF = GetAtomSize( apSINF );

        int [] SampleSizeTable = GetSampleSizeTable();
        int [] ChunkOffsetTable = GetChunkOffsetTable();
        int [] ChunkSamples = GetChunkSamplesTable( ChunkOffsetTable.Length );

        byte [] adUSER = GetAtomData( GetAtomPos( AtomUSER, true ), true );
        byte [] adKEY = GetAtomData( GetAtomPos( AtomKEY, true ), true );
        byte [] adIVIV = GetAtomData( GetAtomPos( AtomIVIV, true ), false );
        byte [] adNAME = GetAtomData( GetAtomPos( AtomNAME, true ), false );
        byte [] adPRIV = GetAtomData( GetAtomPos( AtomPRIV, true ), false );

        uint UserID = BitConverter.ToUInt32( adUSER, 0 );
        uint KeyID = BitConverter.ToUInt32( adKEY, 0 );
        string strName = Encoding.ASCII.GetString( adNAME );
        byte [] UserKey = GetUserKey( UserID, KeyID );

        MD5CryptoServiceProvider MD5 = new MD5CryptoServiceProvider();
        MD5.TransformBlock( adNAME, 0, strName.IndexOf( '\0' ), adNAME, 0 );
        MD5.TransformFinalBlock( adIVIV, 0, adIVIV.Length );

        Decrypt( adPRIV, 0, adPRIV.Length, UserKey, MD5.Hash );

        if( Encoding.ASCII.GetString( adPRIV, 0, 4 ) != "itun" )
        {
            throw new Exception( "Decryption of 'priv' atom failed" );
        }

        Buffer.BlockCopy( adPRIV, 24, Key, 0, Key.Length );
        Buffer.BlockCopy( adPRIV, 48, IV, 0, IV.Length );

        for( int Chunk = 0, Sample = 0;
             Chunk < ChunkOffsetTable.Length; Chunk++ )
        {
            int Pos = ChunkOffsetTable[ Chunk ];

            for( int i = 0; i < ChunkSamples[ Chunk ]; i++ )
            {
                Decrypt( sbuffer, Pos, SampleSizeTable[ Sample ], Key, IV );

                Pos += SampleSizeTable[ Sample++ ];
            }
        }

        Encoding.ASCII.GetBytes( AtomMP4A ).CopyTo( sbuffer, apDRMS );
        Encoding.ASCII.GetBytes( AtomSINF.ToUpper() ).CopyTo( sbuffer, apSINF );
        LeReverse( BitConverter.GetBytes( asDRMS - asSINF ), 0, 1 ).CopyTo(
            sbuffer, apDRMS - 4 );
        Array.Reverse( sbuffer, apGEID, 4 );

        bw.Seek( 0, SeekOrigin.Begin );
        bw.Write( sbuffer );
    }
}

class DeDRMS
{
    // CHANGED(1): iOpener
	// public static void Main( string [] Args )
	public static void Decrypt( string [] Args )
    {
        FileStream fs = null;

        if( Args.Length != 1 )
        {
            Console.WriteLine( "Usage: DeDRMS file.m4p" );
            return;
        }

		try
		{
			fs = new FileStream( Args[ 0 ], FileMode.Open,
				FileAccess.Read | FileAccess.Write );
			M4PStream m4p = new M4PStream( fs );
			m4p.DeDRMS();
		}
		// ADDED(3): iOpener
		catch(iOpener.KeyNotFoundException knf) 
		{
			throw;
		}
		catch( Exception e )
		{
			Console.WriteLine( "Exception: {0}", e.Message );
		}
		finally
		{
			if( fs != null )
			{
				fs.Close();
			}
		}
    }
}
